"""

Plugin Instance managements

This was separated from the plugin module to keep it shorter.  Here we keep
tools for centrally managing plugins and lists of plugins.

"""
import logging
from typing import List

logger = logging.getLogger('configerus.instance')


class PluginInstances:
    """ List of plugins wrapped in the PluginInstance struct so that it can be
        prioritized and managed.

        Question: Why take a factory/copy function instead of just building here?
        Answer: this functionality ended up being useful as an export but was
            limited in reusability without dynamic construction and copying
    """

    def __init__(self, plugin_factory):
        """
        Parameters:
        -----------

        plugin_factory (Callable) : function to product new plugin objects on
            request.

            signature is:
                plugin_factory(type, plugin_id, instance_id, priority)

        """
        self.instances = []
        """ keep a list of all of the plugins as PluginInstance wrappers

            This mixes plugin types together but but it simplifies management
            and storage of plugins """

        self.plugin_factory = plugin_factory
        """ A factory method to produce new plugins """

    def copy(self, new_plugin_factory, plugin_copier):
        """ Make a copy of this plugin list.

        This copies the instance list and by copying every plugin, overriding
        its config and adding it to the new list

        Parameters:
        -----------

        new_plugin_factory (Callable) : a new callable to replace the one for this
            instance.

        plugin_copier (Callable) : a callable which should accept the plugin
            object, and produce a copy of it.

        Returns:
        --------

        A copy of this InstanceList, with copies of all of the PluginInstance
        instances with copies of the plugins.
        Changing/Using the copy should not affect the original.

        """
        instances_copy = PluginInstances(new_plugin_factory)
        """A new instance list which we will return after copying over plugins """

        for instance in self.instances:
            plugin_copy = plugin_copier(instance.plugin)
            instances_copy.instances.append(
                PluginInstance(
                    instance.type,
                    instance.plugin_id,
                    instance.instance_id,
                    instance.priority,
                    plugin_copy))
        return instances_copy

    def add_plugin(self, type: str, plugin_id: str, instance_id: str, priority: int) -> object:
        """ Create a plugin and add it to the list

        Parameters:
        -----------

        type (str) : plugin type as a string (used for filtering)

        plugin_id (str) : id of the plugin as registered using the plugin
            factory decorater. This has to match a plugin's registration with

        instance_id (str) : Optionally give a plugin instance a name, which it
            might use for internal functionality.

        priority (int) : plugin priority. Use this to set this plugins as
            higher or lower priority than others.

        Returns:
        --------

        Returns the plugin so that you can do any actions to the plugin that it
        supports, and the code here doesn't need to get fancy with function
        arguments

        """

        if not plugin_id:
            raise KeyError("Could not create a plugin as an invalid plugin_id was given: '{}'".format(plugin_id))

        if not type:
            raise KeyError("Could not create a plugin as an invalid type was given: '{}'".format(plugin_id))

        if not instance_id:
            # generate some kind of unique instance_id key
            base_instance_id = "{}_{}".format(type.value, plugin_id)
            index = 1
            instance_id = "{}_{}".format(base_instance_id, index)
            while self.has_plugin(instance_id):
                index = 1
                instance_id = "{}_{}".format(base_instance_id, index)

        if self.has_plugin(instance_id, plugin_id):
            self.warn("Replacing '{}.{}' with new plugin instance".format(plugin_id, instance_id))

        try:
            # ask our factory to create the new plugin
            plugin_factory = self.plugin_factory
            plugin = plugin_factory(type=type, plugin_id=plugin_id, instance_id=instance_id, priority=priority)
            instance = PluginInstance(type, plugin_id, instance_id, priority, plugin)

        except NotImplementedError as e:
            raise NotImplementedError(
                "Could not create configerus plugin '{}:{}' as that plugin_id could not be found.".format(
                    type.value, plugin_id)) from e

        self.instances.append(instance)
        return plugin

    def __len__(self) -> int:
        """ Return how many plugin instances we have """
        return len(self.instances)

    def __getitem__(self, instance_id: str) -> object:
        """ Handle subscription request

        For subscriptions assume that an instance_id is being retrieved and that
        a plugin is desired for return.

        Parameters:
        -----------

        instance_id (str) : Instance instance_id to look for

        Returns:

        Plugin object for highest priority plugin with the matching instance_id

        Raises:
        -------

        KeyError if the key cannot be matched,

        """
        return self.get_plugin(instance_id=instance_id)

    """ Accessing plugins """

    def get_plugin(self, plugin_id: str = '', instance_id: str = '',
                   type: str = '', exception_if_missing: bool = True) -> object:
        """ Retrieve the highest priority matching plugin

        Parameters:
        -----------

        instance_id (str) : plugin Instance intance_id for matching
        plugin_id (str) : plugin Instance plugin_id for matching
        type (Type) : plugin Instance type for matching

        All filter parameters are optional.  Asking for none retrieves just the
        highest priority plugin.

        Returns:
        --------

        Highest priority matching plugin object

        """
        instance = self.get_instance(plugin_id=plugin_id, instance_id=instance_id, type=type,
                                     exception_if_missing=exception_if_missing)

        if instance is not None:
            return instance.plugin

    def get_plugins(self, plugin_id: str = '', instance_id: str = '',
                    type: str = '', exception_if_missing: bool = True) -> object:
        """ Retrieve matching plugins

        Parameters:
        -----------

        instance_id (str) : plugin Instance intance_id for matching
        plugin_id (str) : plugin Instance plugin_id for matching
        type (Type) : plugin Instance type for matching

        All filter parameters are optional.  Asking for none retrieves just the
        highest priority plugin.

        Returns:
        --------

        List of sorted matching plugin objects

        """
        instances = self.get_instances(plugin_id=plugin_id, instance_id=instance_id, type=type)

        if not instances:
            if exception_if_missing:
                raise KeyError("Could not find any matching plugins[{}][{}][{}]".format(
                    plugin_id, instance_id, type))
            return []

        return [instance.plugin for instance in instances]

    def has_plugin(self, plugin_id: str = '', instance_id: str = '', type: str = '') -> bool:
        """ does a matching plugin exist

        Parameters:
        -----------

        instance_id (str) : plugin Instance intance_id for matching
        plugin_id (str) : plugin Instance plugin_id for matching
        type (Type) : plugin Instance type for matching

        All filter parameters are optional.  Asking for none retrieves just the
        highest priority plugin.

        Returns:
        --------

        bool if a plugin match can be found

        """
        return bool(len(self.get_instances(plugin_id=plugin_id, instance_id=instance_id, type=type)))

    def get_instance(self, plugin_id: str = '', instance_id: str = '',
                     type: str = '', exception_if_missing: bool = True) -> 'PluginInstance':
        """ Retrieve amatching instance

        Parameters:
        -----------

        instance_id (str) : plugin Instance intance_id for matching
        plugin_id (str) : plugin Instance plugin_id for matching
        type (Type) : plugin Instance type for matching

        All filter parameters are optional.  Asking for none retrieves just the
        highest priority plugin.

        Returns:
        --------

        the highest priority matching Instance object

        """
        try:
            sorted_and_filtered = self.get_instances(plugin_id=plugin_id, instance_id=instance_id, type=type)
            return sorted_and_filtered[0]
        except (IndexError, KeyError) as e:
            if exception_if_missing:
                raise KeyError(
                    "No instances matched filter requirements [{}][{}][{}]".format(
                        plugin_id, instance_id, type)) from e
            else:
                logger.debug("No instances matched for filter requirements")

    def get_instances(self, plugin_id: str = '', instance_id: str = '',
                      type: str = '') -> List['PluginInstance']:
        """ filter and order plugins by their top down priority

        Parameters:
        -----------

        instance_id (str) : plugin Instance intance_id for matching
        plugin_id (str) : plugin Instance plugin_id for matching
        type (Type) : plugin Instance type for matching

        All filter parameters are optional.  Asking for none retrieves just the
        highest priority plugin.

        Returns:
        --------

        List of filtered and sorted PluginInstance objects

        """
        matched_instances = []
        for plugin_instance in self.instances:
            if plugin_id and not plugin_instance.plugin_id == plugin_id:
                continue
            elif instance_id and not plugin_instance.instance_id == instance_id:
                continue
            elif type and not type == plugin_instance.type:
                continue

            matched_instances.append(plugin_instance)

        if not len(matched_instances):
            # no need to put more work into it if we have no instances
            return []

        # sort from high to low (avoid divide by zero)
        sorted_instances = sorted(matched_instances, key=lambda instance: 1 /
                                  instance.priority if instance.priority > 1 else 1)
        return sorted_instances


class PluginInstance:
    """ a struct for a plugin instance that also keeps metadata about the instance """

    def __init__(self, type: str, plugin_id: str, instance_id: str, priority: int, plugin):
        self.type = type
        self.plugin_id = plugin_id
        self.instance_id = instance_id
        self.priority = priority
        self.plugin = plugin



import logging
from importlib import metadata
import copy
from typing import Any

from .plugin import Factory, Type
from .instances import PluginInstances
from .shared import tree_merge
from .loaded import Loaded
from .validator import ValidationError

logger = logging.getLogger('configerus:config')

CONFIGERUS_BOOTSTRAP_ENTRYPOINT = "configerus.bootstrap"
""" SetupTools entrypoint target for bootstrapping """

CONFIG_PATH_LABEL = 'paths'
""" If you load this label, it is meant to be return a keyed path """

PLUGIN_DEFAULT_PRIORITY = 75
""" Default plugin priority, which should be a common unprioritized value

General approach:

<35 Low priority defaults (system)
<50 Higher priority defaults (contrib)
<75 Low prority setting (project)
 75 default
<90 High priority settings (project, contrib)
>90 !important (project)

"""


class Config:
    """ Config management class (v3)

    Allows some sourced file based project configuration with some pattern
    templating and path set overriding.

    To allow easy separation of config into manageable files, all config is
    grouped into "label" strings which correlate to config files with matching
    names.
    Before accessing a config label, it has to be loaded, which you can do with
    the load() method.
    Then all get() calls use the most recently loaded config label.

    Templating: some string substitution can be done based on config values
        and some path keys. Templating options are based on a custom syntax

    Overriding: tell config about multiple paths in order of priority, and it
        will allow higher priority values to override lower priority values

    Dot notation retrival: you can treat your data as a tree, and config will
        allow you to retrieve tree nodes using a syntax where "." is a node
        label delimiter.

    """

    def __init__(self):
        self.plugins = PluginInstances(self.make_plugin)
        """ keep a list of all of the plugins as PluginInstance wrappers

            This mixes plugin types together but but it simplifies management
            and storage of plugins """

        self.loaded = {}
        """ cache of config that has been loaded """

    def copy(self):
        """ make a copy of this config object so that we can independently edit
            the copy without affecting the original.
        """
        copy = Config()
        copy.plugins = self.plugins.copy(copy.make_plugin, copy.copy_plugin)
        return copy

    def bootstrap(self, bootstrap_id: str):
        """
        Make a plugin object instance of a type and key

        A python module out there will need to have declared an entry_point for
        'mirantis.testing.toolbox.{type}' with entry_point key {name}.
        The entrypoint must be a factory method of signature:
        ```
            def XXXX(conf: configerus.config.Config, instance_id:str) -> {plugin}:
        ```

        An alternative would be if the entrypoint refers to a sub-package here. We
        directly

        The factory should return the plugin, which is likely going to be some kind
        of an object which has value to the caller of the function. This function
        does not specify what the return needs to be.

        """
        logger.info("Running configerus bootstrap entrypoint: %s", bootstrap_id)
        eps = metadata.entry_points()[CONFIGERUS_BOOTSTRAP_ENTRYPOINT]
        for ep in eps:
            if ep.name == bootstrap_id:
                bootstrap_entrypoint = ep.load()
                bootstrap_entrypoint(self)
                break
        else:
            raise KeyError("Bootstrap not found {}:{}".format(CONFIGERUS_BOOTSTRAP_ENTRYPOINT, bootstrap_id))

    """ Pass through accessors

    These methods just give access to some values used to govern config without
    having to import the module and directly access constants.

    """

    def paths_label(self):
        """ retrieve the special config label which is a list of paths """
        return CONFIG_PATH_LABEL

    def default_priority(self):
        """ Return the default priority for relative priority setting """
        return PLUGIN_DEFAULT_PRIORITY

    """ Plugin factory """

    def make_plugin(self, type: Type, plugin_id: str, instance_id: str, priority: int = PLUGIN_DEFAULT_PRIORITY):
        """ Make a new plugin """
        fac = Factory(type, plugin_id)
        return fac.create(self, instance_id)

    def copy_plugin(self, plugin: object):
        """ Make a plugin object """
        plugin_copy = copy.deepcopy(plugin)
        plugin_copy.config = self
        return plugin_copy

    """ Source plugins

    All methods related to managing and using source plugins.

    Some methods are short wrappers on the plugin abstractions, but that allows
    consumers to not have to import the .plugin.Type enums

    """

    def has_source(self, instance_id: str):
        """ Check if a source instance has already been added """
        return self.plugins.has_plugin(instance_id, Type.SOURCE)

    def add_source(self, plugin_id: str, instance_id: str = '', priority: int = PLUGIN_DEFAULT_PRIORITY):
        """ add a new config source to the config object and return it

        Parameters:
        -----------

        plugin_id (str) : id of the plugin as registered using the plugin factory
            decorater for a source plugin.  This has to match a plugin's registration
            with the plugin factory as a part of the Factory decoration

        instance_id (str) : Optionally give a source plugin instance a name, which
            it might use for internal functionality.
            The "path" source plugin allows string template substitution of the
            "__paths__:instance_id" for the path.

        priority (int) : source priority. Use this to set this source values as
            higher or lower priority than others.

        Returns:
        --------

        Returns the source plugin so that you can do any actions to the plugin that it
        supports, and the code here doesn't need to get fancy with function arguments

        """
        # drop any loaded config
        self.loaded = {}
        # add the plugin to our list.
        return self.plugins.add_plugin(Type.SOURCE, plugin_id, instance_id, priority)

    def load(self, label: str, force_reload: bool = False, validator: str = ""):
        """ Load a config label

        This queries all sources for the label and merges the loaded config data
        into a single Loaded object.
        The loads are cached per Config instance to make repeated calls cheap.

        Parameters:
        -----------

        label (str) : Config label to load.  Every config source can provide
            config data sets for labels, which correspond to things like
            "all files with a matching filename", or all "records in a matching
            db table"

        force_reload (bool) : if true, and data has been loaded before, this
            forces a fresh reload of data from the sources

        validator (str) : string key passed to the validate() method which will
            validate the retrieved data before returning it.

        Returns:
        --------

        A Loaded object from which you can .get() specific pieces of config

        Throws:
        -------

        Some source plugins will throw Exceptions if they have internal problems

        If you request a validation, and validation fails, then a Validation
        error with be raised.

        """
        if force_reload or label not in self.loaded:
            """ load data from all of the sources for a label """
            logger.debug("Loading Config '%s' from all sources", label)

            data = {}
            # merge in data from the higher priorty into the lower priority
            for source in self.plugins.get_plugins(type=Type.SOURCE):
                source_data = source.load(label)
                if source_data:
                    data = tree_merge(data, source_data)

            if not data:
                raise KeyError(
                    "Config '{}' loaded data came out empty.  That means that no config source could find that label.  That is likely a problem".format(label))

            self.loaded[label] = Loaded(data=data, parent=self, instance_id=label)

        if validator:
            self.validate(self.loaded[label].data, validator)

        logger.debug("Loaded config {} : {}".format(label, self.loaded[label].data))
        return self.loaded[label]

    """ Formatter plugin usage and management """

    def has_formatter(self, instance_id: str):
        """ Check if a formatter instance has already been added

        You can use this in abstracts to detect if you've already added a plugin
        """
        return self.plugns.has_plugin(instance_id, Type.FORMATTER)

    def add_formatter(self, plugin_id: str, instance_id: str = '', priority: int = PLUGIN_DEFAULT_PRIORITY):
        """ add a new config formatter to the config object and return it

        Parameters:
        -----------

        plugin_id (str) : id of the plugin as registered using the plugin
            factory  decorater for a formatter plugin.  This has to match a
            plugin's registration  with the plugin factory as a part of the
            Factory decoration

        instance_id (str) : Optionally give a formatters plugin instance a name,
            which it might use for internal functionality.

        priority (int) : source priority. Use this to set this source values as
            higher or lower priority than others.

        Returns:
        --------

        Returns the source plugin so that you can do any actions to the plugin that it
        supports, and the code here doesn't need to get fancy with function arguments

        """
        return self.plugins.add_plugin(Type.FORMATTER, plugin_id, instance_id, priority)

    def format(self, data, default_label: Any, validator: str = ""):
        """ Format some data using the config object formatters

        data (Any): primitive data that should be formatted. The data will be
            passed to the formatter plugins in descending priority order.

        default_label (str) : because string templating can indicate a replace
            target without a label source, a default source is used to tell a
            formatter what config label to use in its absence.

        validate_target (Any) : Validation target which will be interpreted by the
            validation plugins.  Two typical formats are common:

            1. a string with some identifiable pattern that a plugi can recognize
            2. a dict with keys that different plugins can recognize.

            if empty/None then no validation is performed.

        """
        for formatter in self.plugins.get_plugins(type=Type.FORMATTER, exception_if_missing=False):
            data = formatter.format(data, default_label)

        if validator:
            self.validate(self.loaded[label].data, validator)

        return data

    """ Validator plugin usage and management """

    def has_validator(self, instance_id: str):
        """ Check if a formatter instance has already been added

        You can use this in abstracts to detect if you've already added a plugin
        """
        return self.plugns.has_plugin(instance_id, Type.VALIDATOR)

    def add_validator(self, plugin_id: str, instance_id: str = '', priority: int = PLUGIN_DEFAULT_PRIORITY):
        """ add a new config validator to the config object and return it

        Parameters:
        -----------

        plugin_id (str) : id of the plugin as registered using the plugin
            factory  decorater for a validator plugin.  This has to match a
            plugin's registration  with the plugin factory as a part of the
            Factory decoration

        instance_id (str) : Optionally give a validator plugin instance a name,
            which it might use for internal functionality.

        priority (int) : source priority. Use this to set this source values as
            higher or lower priority than others.

        Returns:
        --------

        Returns the source plugin so that you can do any actions to the plugin that it
        supports, and the code here doesn't need to get fancy with function arguments

        """
        return self.plugins.add_plugin(Type.VALIDATOR, plugin_id, instance_id, priority)

    def validate(self, data, validate_target: str, exception_if_invalid: bool = True):
        """ Format some data using the config object formatters

        data (Any): primitive data that should be formatted. The data will be
            passed to the formatter plugins in descending priority order.

        validate_target : A config key which can be used

        Returns:
        --------

        Bool results of validation. If any validation plugin raises an exception
        then it will be false.

        If exception_if_invalid then validation failures with raise and exception

        Raises:
        -------

        ValidationError on any validation exception.

        ValdiationError if any validator ran into an exception internally.

        @TODO can we be more specific in interpreting validation exceptions.

        """

        try:
            for validator in self.plugins.get_plugins(type=Type.VALIDATOR, exception_if_missing=False):
                validator.validate(validate_target, data)
        except Exception as e:
            if exception_if_invalid:
                raise ValidationError("Config validation failed: {}".format(e)) from e
            return False
        return True



import logging

from configerus.config import Config
from configerus.plugin import SourceFactory

from .source import ConfigSourceEnvPlugin

PLUGIN_ID_SOURCE_ENV = 'env'
""" ConfigSource plugin_id for the configerus env configsource plugin """
CONFIGERUS_ENV_BASE_KEY = 'base'
""" Config key for retrieving an env plugin base value from config """

@SourceFactory(plugin_id=PLUGIN_ID_SOURCE_ENV)
def plugin_factory_configsource_env(config: Config, instance_id: str = ''):
    """ create an configsource env plugin """
    return ConfigSourceEnvPlugin(config, instance_id)


def configerus_bootstrap(config: Config):
    """ Bootstrap a config object

    We don't actually do anything, so this bootstrapper is here only to ensure
    that the above factory decorator is run

    """
    pass




from configerus.config import Config
from configerus.plugin import FormatFactory

from .format import ConfigFormatDeepPlugin

PLUGIN_ID_FORMAT_DEEP = 'deep'
PLUGIN_PRIORITY_FORMAT_DEEP = 90

""" Format plugin_id for the configerus deep format plugin """


@FormatFactory(plugin_id=PLUGIN_ID_FORMAT_DEEP)
def plugin_factory_format_deep(config: Config, instance_id: str = ''):
    """ create an format plugin which recursiveley formats data """
    return ConfigFormatDeepPlugin(config, instance_id)


def configerus_bootstrap(config: Config):
    """ Bootstrap a config object by adding our formatter """
    config.add_formatter(PLUGIN_ID_FORMAT_DEEP, priority=PLUGIN_PRIORITY_FORMAT_DEEP)



import logging

from configerus.config import Config
from configerus.plugin import SourceFactory

from .source import ConfigSourceDictPlugin

PLUGIN_ID_SOURCE_DICT = 'dict'
""" ConfigSource plugin_id for the configerus dict configsource plugin """


CONFIGERUS_DICT_DATA_KEY = 'data'
""" if you load paths from config, this is the key that should give the path """

@SourceFactory(plugin_id=PLUGIN_ID_SOURCE_DICT)
def plugin_factory_configsource_dict(config: Config, instance_id: str = ''):
    """ create an configsource dict plugin """
    return ConfigSourceDictPlugin(config, instance_id)


def configerus_bootstrap(config: Config):
    """ Bootstrap a config object

    We don't actually do anything, so this bootstrapper is here only to ensure
    that the above factory decorator is run

    """
    pass



import os
import re
import json
import yaml
import logging

from configerus.config import Config
from configerus.shared import tree_merge

logger = logging.getLogger('configerus.contrib.source')

# FileTypes that this class can use at this time
FILESOURCE_FILETYPES = ["json", "yaml", "yml"]
""" Valid config types that the config loader can currently handled """

CONFIGERUS_PATH_LABEL = 'paths'
""" If you load this label, it is meant to be return a keyed path """


class ConfigSourcePathPlugin:
    """   """

    def __init__(self, config: Config, instance_id: str):
        """  """
        self.config = config
        self.instance_id = instance_id

        self.path = ''

    def set_path(self, path: str):
        """ Set the config path source """
        self.path = path

    def load(self, label: str):
        """ load config for a name

        Parameters:
        -----------

        lable (str) : config label to load, should correlated to a json or yaml
            file of the same name in the path, otherwise an empty Dict is
            returned.

            **There is 1 special case, where if MTT_CONFIG_PATH_LABEL is passed
              then the function returns a Dict of 'instance_id:path' which can
              be used for string substitution**

        Returns:
        --------

        Dict[str, Any] of data that was loaded for the label
        """

        if not os.path.isdir(self.path):
            raise ValueError(
                "Could not load '{}' path config, as the source path does not exist: {}".format(
                    self.instance_id, self.path))

        # Special case for retreiving paths instead of config
        if label == CONFIGERUS_PATH_LABEL:
            return {self.instance_id: self.path}

        data = {}
        """ hold all merged data from found source files """

        file_types_re = "|".join(FILESOURCE_FILETYPES)
        """ regex part for allowed file extensions """
        config_files_re = rf"({label})\.({file_types_re})"
        """ regex that matches all valid config filenames for the label """

        for file in [f for f in os.listdir(self.path) if re.match(config_files_re, f)]:
            with open(os.path.join(self.path, file)) as matching_file:
                extension = os.path.splitext(file)[1].lower()
                if extension == ".json":
                    try:
                        file_config = json.load(matching_file)
                    except json.decoder.JSONDecodeError as e:
                        raise ValueError(
                            "Failed to parse one of the config files '{}': {}".format(
                                os.path.join(
                                    self.path, file), e))

                    assert file_config, "Empty config in {} from file {}".format(path, file)
                elif extension == ".yml" or extension == ".yaml":
                    try:
                        file_config = yaml.load(matching_file, Loader=yaml.FullLoader)
                    except yaml.YAMLError as e:
                        raise ValueError(
                            "Failed to parse one of the config files '{}': {}".format(
                                os.path.join(
                                    self.path, file), e))

                    assert file_config, "Empty config in {} [{}]".format(file, self.path)

                else:
                    raise ValueError(
                        "Unknown config filetype. Cannot parse '{}' files, but it matches our regex".format(extension))

                data = tree_merge(file_config, data)

        return data



"""

Test config validation

Here we run some simple config validation on both load() and get() operations
using the configerus.contric.jsonscheme validator plugin.

"""

from configerus.contrib.jsonschema import PLUGIN_ID_VALIDATE_JSONSCHEMA
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT
import configerus
import unittest
import logging

logging.basicConfig()
logger = logging.getLogger('validate')
logger.setLevel(level=logging.INFO)


""" TESTS """


class JsonSchemaValidate(unittest.TestCase):
    """ Some simple validation testing usinf jsonschema """

    # examples from https://python-jsonschema.readthedocs.io/en/stable/
    valid_instance = {"name": "Eggs", "price": 34.99}
    invalid_instance = {"name": "Eggs", "price": "Invalid"}
    instance_schema = {
        "type": "object",
        "properties": {
            "price": {"type": "number"},
            "name": {"type": "string"},
        }
    }

    def _simple_validate_config(self):
        """ use a common config across some tests """

        logger.debug("Building jsonschema config object")
        config = configerus.new_config()
        config.add_validator(PLUGIN_ID_VALIDATE_JSONSCHEMA)
        config.add_source(PLUGIN_ID_SOURCE_DICT).set_data({
            'valid_load_test': self.valid_instance,
            'invalid_load_test': self.invalid_instance,
            'get_test': {
                'valid': self.valid_instance,
                '1': {
                    'invalid': self.invalid_instance
                }
            },
            'jsonschema': {
                'instance': self.instance_schema
            }
        })

        return config

    def test_validate_jsonschema_load_valid(self):
        """ just do a minimal jsonschema validation """

        config = self._simple_validate_config()
        load_config = config.load('valid_load_test', validator='jsonschema:instance')

    def test_validate_jsonschema_load_invalid(self):
        """ just do a minimal jsonschema validation """

        config = self._simple_validate_config()
        with self.assertRaises(Exception):
            config.load('invalid_load_test', validator='jsonschema:instance')

    def test_validate_jsonschema_get(self):
        """ just do a minimal jsonschema validation """

        config = self._simple_validate_config()
        test_config = config.load('get_test')

        # this should be valid
        valid_instance = test_config.get('valid', validator='jsonschema:instance')

        with self.assertRaises(Exception):
            test_config.get('1.invalid', validator='jsonschema:instance')



from typing import Dict, Any
import logging
import os

from configerus.config import Config

logger = logging.getLogger('configerus.contrib.env.source')


class ConfigSourceEnvPlugin():
    """ Get config from ENV variables

    if a base is provided, it is used as a root for all ENV variables
    as "{base}_"

    string case for keys is a weird topic.  We .lower() all keys as it seems a
    decent convention.

    """

    def __init__(self, config: Config, instance_id: str):
        """  """
        self.config = config
        self.instance_id = instance_id

        self.base = ''
        """ keep the ENV base prefix that should limit what gets pulled in """

    def set_base(self, base: str):
        self.base = base

    def load(self, label: str):
        """ Load a config label and return a Dict[str, Any] of config data

        Parameters:

        label (str) : label to load

        """

        if self.base:
            label_prefix = '{}_{}_'.format(self.base, label).upper()
        else:
            label_prefix = '{}_'.format(label).upper()

        filtered = {}
        for env_key, env_value in os.environ.items():
            if env_key.upper().startswith(label_prefix):
                filtered[env_key[len(label_prefix):].lower()] = env_value

        organized = {}
        for (key, value) in filtered.items():
            if key.startswith('_'):
                continue

            base = organized
            key_steps = key.lower().split('_')

            last_step = key_steps.pop()

            for key_step in key_steps:
                if not isinstance(base, dict):
                    base = {}
                if not key_step in base:
                    base[key_step] = {}
                base = base[key_step]
            if not isinstance(base, dict):
                base = {}
            base[last_step] = value

        return organized



from typing import Any
from jsonschema import validate

from configerus.config import Config

PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL = 'jsonschema'


class JsonSchemaValidatorPlugin():
    """   """

    def __init__(self, config: Config, instance_id: str):
        """  """
        self.config = config
        self.instance_id = instance_id

    def validate(self, validate_target: Any, data):
        """ Validate a structure using jsonschema pulled from a config key

        Parameters:
        -----------

        validate_target (str|dict[str:dict]) : validation schema indicatior:

            One of two options will be validated, otherwise we will ignore:

            1. A string of the format "jsonschema:{validation-config-key}"
            2. dict with key "jsonschema"

            For #1, the validation target is expected to be a config target
            (label:key) which will retrieve jsonschema schema from config

            For #2 a dict value for key 'jsonschema' is expected to be the
            schema.

        data (Any) : data which should be validated

        Raises:
        -------

        If a jsonchema schema was identified in the validate target, then a
        jsonschema validation error will be raised if the data is not valid.

        """

        if isinstance(validate_target, str):
            (method, validate_key) = validate_target.split(':')

            if not method == PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL:
                # returns of any value signal validation as validate only catches exceptions
                return

            try:
                schema_config = self.config.load(PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL)
                schema = schema_config.get(validate_key, exception_if_missing=True)

            except Exception as e:
                raise NotImplementedError("Could not access jsonschema validation schema from config target '{}:{}'".format(PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL, validate_key))

        elif isinstance(validate_target, dict):
            if not PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL in validate_target:
                return

            schema = validate_target[PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL]

            if not isinstance(schema, dict):
                raise ValueError("JSONSCHEMA scheme was expected to be a dict: {}".format(schema))

        else:
            # Could not interpret validate target
            return

        validate(data, schema=schema)



import re
import logging

from configerus.config import Config

logger = logging.getLogger('configerus.contrib.get:formatter')

CONFIG_DEFAULT_MATCH_PATTERN = r'\{((?P<source>\w+):)?(?P<key>[-\w.]+)(\?(?P<default>[^\}]+))?\}'
""" Default regex pattern used to string template, which needs to identify
keys that can be used for replacement. Needs to have a named group for
key, and can have named group for source and default

The above regex pattern resolves for three named groups: source, key and default:

the {first?check} value          : key=first, default=check
a {common.dot.notation} test     : key=commond.dot.notation
a {label:dot.notation.test} here : source=label, key=dot.notation.test

so:
1. all replacements are wrapped in "{}"
2. an optional "source:" group tells config to look in a specifig label
3. a "?default" group allows a default (untemplated) value to be used if the
   key cannot be found
"""


class ConfigFormatGetPlugin():
    """   """

    def __init__(self, config: Config, instance_id: str):
        """  """
        self.config = config
        self.instance_id = instance_id

        self.pattern = re.compile(CONFIG_DEFAULT_MATCH_PATTERN)

    def format(self, target, default_source: str):
        """ Format a string by substituting config values

        target: a string that should be formatted. If not a string then no
            formatting is performed

        default_source : if format/replace patterns don't have a source defined
            then this is used as a source.

        """

        if not isinstance(target, str):
            return target

        # if the entire target is the match, then replace whatever type we get
        # out of the config .get() call
        match = self.pattern.fullmatch(target)
        if match:
            return self._get(match, default_source)

        # search through the target replacing any found matches with
        start = 0
        match = self.pattern.search(target, start)
        while match:
            rep = str(self._get(match, default_source))
            target = target[:match.start()] + rep + target[match.end():]
            start = start + len(rep)
            match = self.pattern.search(target, start)

        return target

    def _get(self, match, default_source: str):
        """ from an re.match get data from config """
        key = match.group('key')
        source = match.group('source')
        default = match.group('default')

        if source is None:
            source = default_source
        source_config = self.config.load(source)
        try:
            return source_config.get(key, exception_if_missing=True)
        except KeyError as e:
            if match.group('default'):
                return match.group('default')
            else:
                # if a template string wasn't found then exception
                raise e



"""
test_config

Unit and functional testing for the config handler

Here we test that config can load() and get() config and that those operations
behave as expected.
To get some good coverage we try to use both the DICT and PATH based config
plugins.  To do that we define a list of config and use a configerus.test
utility method which turns the PATH content into files in a temp directory
before adding it to the config object.
This makes this test harder to read, but it is a clean way to get better tests.

"""
import logging
import unittest

import configerus
from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.env import PLUGIN_ID_SOURCE_ENV
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT
from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH

from configerus.test import make_test_config, test_config_cleanup

logger = logging.getLogger("test_config_behaviour")
logger.setLevel(logging.INFO)

config_sources = [
    {
        'name': 'first',
        'priority': 30,
        'type': PLUGIN_ID_SOURCE_DICT,
        'data': {
            'config': {
                '1': "first 1"
            },
            'variables': {
                'one': "first one",
                'two': "first two"
            }
        }
    },
    {
        'name': 'second',
        'priority': 20,
        'type': PLUGIN_ID_SOURCE_PATH,
        'data': {
            'config.json': {
                '1': "second 1",
                '2': "second 2"
            },
            'variables.json': {
                'one': "second one",
                'two': "second two"
            }
        }
    },
    {
        'name': 'third',
        'priority': 40,
        'type': PLUGIN_ID_SOURCE_PATH,
        'data': {
            'config.yaml': {
                '3': {
                    '1': "third 3.1",
                    '2': {
                        '1': "third 3.2.1",
                        '2': "third 3.2.2"
                    }
                },
                '4': {
                    '1': "third 4.1"
                },
                '5': "third 5"
            }
        }
    },
    {
        'name': 'fourth',
        'priority': 75,
        'type': PLUGIN_ID_SOURCE_PATH,
        'data': {
            'config.json': {
                '4': "fourth 4"
            }
        }
    },
    {
        'name': 'fifth',
        'priority': 75,
        'type': PLUGIN_ID_SOURCE_PATH,
        'data': {
            'config.json': {
                '5': "fifth 5",
                '6': "fifth 6"
            }
        }
    },
    {
        'name': 'sixth',
        'priority': 75,
        'type': PLUGIN_ID_SOURCE_PATH,
        'data': {
            'config.json': {
                '6': {
                    '1': "sixth 6.1"
                }
            }
        }
    },
    {
        'name': 'seventh',
        'priority': 85,
        'type': PLUGIN_ID_SOURCE_PATH,
        'data': {
            'config.json': {
                '5': "seventh 5 json",
                '6': "seventh 6 json"
            },
            'config.yaml': {
                '5': "seventh 5 yaml",
                '6': "seventh 6 yaml",
                '7': {
                    '1': ['1']
                }
            }
        }
    },
    {
        'name': 'eighth',
        'priority': 85,
        'type': PLUGIN_ID_SOURCE_ENV,
        'data': {
            'env_one': "env eighth env one",
            'env_two_one': "env eighth env two.one",
        }
    }
]
""" Contents of test config files used as the source for a config object """

""" TESTS """


class ConfigBehaviour(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ make a Config test object from some inline data

        First this dumps the data into json files, and then points the config object
        to the various paths for loading. The data can then be used to test funct.
        The data contains only values that are usefull for confirming that config
        behaviour is as expected, and is not meant to be useful.

        We also preload some of the config
        """

        logger.debug("Building empty config object")
        config = configerus.new_config()
        make_test_config(config, config_sources)
        cls.config = config
        cls.loaded_config = config.load('config')
        cls.loaded_variables = config.load('variables')

    @classmethod
    def tearDownClass(cls):
        test_config_cleanup(cls.config)

    def test_source_env_basic(self):
        """ test the env source basically """
        all_env = self.config.load('env')
        self.assertIsNotNone(all_env.get('one'))
        self.assertEqual(all_env.get('one'), "env eighth env one")
        self.assertEqual(all_env.get('two.one'), "env eighth env two.one")

    def test_basic_combined(self):
        """ test some basic file combining by the config object """

        self.assertEqual(self.loaded_config.get('1'), "first 1")
        self.assertEqual(self.loaded_config.get('2'), "second 2")

    def test_dot_notation(self):
        """ Confirm that we can retrieve data using the dot notation """

        self.assertEqual(self.loaded_config.get('3.1'), "third 3.1")

    def test_overrides(self):
        """ confirm that keys defined in more than one source get overriden """

        self.assertEqual(self.loaded_config.get('4'), "fourth 4")
        self.assertIsNone(self.loaded_config.get('5.1', exception_if_missing=False))
        self.assertEqual(self.loaded_config.get('5'), "seventh 5 json")

    def test_get_multiple_keys(self):
        """ test the the loaded get of various key formats """

        third_3_base = self.loaded_config.get('3')
        third_3_2 = self.loaded_config.get('3.2')
        third_3_2_1 = self.loaded_config.get('3.2.1')

        # sanity check on the comparison values
        self.assertIsNotNone(third_3_base)
        self.assertIsInstance(third_3_base, dict)
        self.assertIsInstance(third_3_2, dict)
        self.assertEqual(third_3_2_1, "third 3.2.1")

        # pull some values with a base
        third_3_base_A = self.loaded_config.get(['3', '2'])
        third_3_base_B = self.loaded_config.get(['3', '2.1'])
        third_3_base_C = self.loaded_config.get(['3', ['2', '1']])
        third_3_base_D = self.loaded_config.get([['3', '2'], '1'])

        self.assertEqual(third_3_2, third_3_base_A)
        self.assertEqual(third_3_2_1, third_3_base_B)
        self.assertEqual(third_3_2_1, third_3_base_C)
        self.assertEqual(third_3_2_1, third_3_base_D)

        # Load with some root keys involved
        third_root_base_3 = self.loaded_config.get([LOADED_KEY_ROOT, '3'])
        third_3_base_root = self.loaded_config.get([LOADED_KEY_ROOT, '3'])

        self.assertEqual(third_3_base, third_root_base_3)
        self.assertEqual(third_3_base, third_3_base_root)



from .config import Config
import logging
from typing import List

logger = logging.getLogger('configerus')


CONFIGERUS_BOOSTRAP_DEFAULT = [
    'deep',
    'get',
    'files'
]
""" Default list of modules to bootstrap for a new config object """


def new_config(bootstraps: List[str] = CONFIGERUS_BOOSTRAP_DEFAULT):
    """ Get a new Config object

    bootstraps (List[str]) : list of modules to bootstrap for this config object
        For each string, config will try to run a setuptools entrypoint bootstrap
        function, passing in the config object.

        This can be used for 2 purposes:
        1. the bootstrap function may be in a file which registers plugins that
            can be used in the config object
        2. the bootstrap can modify the config object as needed.

    """
    # start with a new config object
    config = Config()

    for bootstrap_entrypoint_id in bootstraps:
        logger.debug('bootstrapping: {}'.format(bootstrap_entrypoint_id))
        config.bootstrap(bootstrap_entrypoint_id)

    return config



"""

Test config templating

Here we run some simple config template checking to make sure that template
substitution of whole and partial values works.

"""
import logging
import unittest
import os.path
import json

import configerus
from configerus.config import CONFIG_PATH_LABEL
from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT
from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH

from configerus.test import make_test_config, test_config_cleanup

logger = logging.getLogger('test_config_behaviour')
logger.setLevel(logging.INFO)

config_sources = [
    {
        'name': 'source',
        'priority': 30,
        'type': PLUGIN_ID_SOURCE_DICT,
        'data': {
            'config': {
                '1': "first 1",
                '2': {
                    '1': "first 2.1"
                },
                '3': {
                    '1': {
                        '1': "first 3.1.1",
                        '2': "first 3.1.2"
                    },
                    '2': {
                        '1': "first 3.2.1",
                        '2': "first 3.2.2"
                    }
                }
            },
            'variables': {
                'one': "first one",
                'two': "first two",
                'three': 3,
                'four': {
                    '1': "first four.1"
                }
            }
        }
    },
    {
        'name': 'simple',
        'priority': 80,
        'type': PLUGIN_ID_SOURCE_DICT,
        'data': {
            'config': {
                '10': "{2}",
                '11': "__{2}__"
            }
        }
    },
    {
        'name': 'defaults',
        'priority': 80,
        'type': PLUGIN_ID_SOURCE_DICT,
        'data': {
            'config': {
                '20': "{does.not.exist?default}",
                '21': "{does.not.exist}",
                '22': "{variables:does.not.exist?megadefault}",
                '23': "{variables:three?default}",
            }
        }
    },
    {
        'name': 'special',
        'priority': 80,
        'type': PLUGIN_ID_SOURCE_DICT,
        'data': {
            'config': {
                '30': "{path:replace}",
            }
        }
    },
    {
        'name': 'deep',
        'priority': 80,
        'type': PLUGIN_ID_SOURCE_DICT,
        'data': {
            'config': {
                '40': "deep 40",
                '41': "{40}",
                '42': {
                    '1': "{40}",
                    '2': {
                        '1': "{40}"
                    }
                }
            }
        }
    },
    {
        'name': 'replace',
        'priority': 80,
        'type': PLUGIN_ID_SOURCE_PATH,
        'data': {
            'replace.json': {
                'id': "test"
            }
        }
    }
]
""" Contents of test config files used as the source for a config object """

""" TESTS """


class ConfigTemplating(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ make a Config test object from some inline data

        First this dumps the data into json files, and then points the config object
        to the various paths for loading. The data can then be used to test funct.
        The data contains only values that are usefull for confirming that config
        behaviour is as expected, and is not meant to be useful.

        We also preload some of the config
        """

        logger.debug("Building empty config object")
        config = configerus.new_config()
        make_test_config(config, config_sources)
        cls.config = config

        cls.loaded_config = config.load('config')
        cls.loaded_variables = config.load('variables')

    @classmethod
    def tearDownClass(cls):
        test_config_cleanup(cls.config)

    def test_format_sane(self):
        """ simple template replacing using format """

        self.assertEqual(self.config.format('{1}', 'config'), "first 1")
        self.assertEqual(self.loaded_config.format('{1}'), "first 1")
        """ Simple formatting at the config and loadedconfig level """

    def test_format_dot_notation(self):
        """ test that format catches dot notation """

        self.assertEqual(self.config.format('{2.1}', 'config'), "first 2.1")
        self.assertEqual(self.config.format('{3.2.1}', 'config'), "first 3.2.1")
        """ Simple dot notation descending format """

    def test_format_embedding_and_casting(self):
        """ simple template replacing using format """

        self.assertEqual(self.config.format('12345{1}67890', 'config'), "12345first 167890")
        self.assertEqual(self.config.format('1 23 45 {1}6 78 90', 'config'), "1 23 45 first 16 78 90")
        """ format replacements in the middle of a string """

        self.assertEqual(self.config.format('{2}', 'config'), {'1': "first 2.1"})
        self.assertEqual(self.config.format(' {2} ', 'config'), " {'1': 'first 2.1'} ")
        """ format should string cast only not full matches """

    def test_format_with_sources(self):
        """ test that sources work with format """

        variables_two = self.config.format("{two}", 'variables')
        self.assertEqual(variables_two, "first two")
        self.assertEqual(self.config.format("{variables:two}", 'config'), variables_two)
        self.assertEqual(self.config.format("{variables:two}", 'variables'), variables_two)
        """ specifying a source should work """

    def test_format_with_defaults(self):
        """ test the defaulting with formatting """

        self.assertEqual(self.config.format("{does.not.exist?default1}", 'config'), 'default1')
        self.assertEqual(self.config.format("{variables:does.not.exist?default2}", 'config'), 'default2')
        """ defaults should apply """

        variables_two = self.config.format("{two}", 'variables')
        self.assertEqual(variables_two, "first two")
        self.assertEqual(self.config.format("{variables:two?default}", 'config'), variables_two)
        """ defaults should not overwrite values that can be found """

    def test_get_formatted(self):
        """ test that loadedconfig gets format """

        config_one = self.loaded_config.get('1')
        self.assertEqual(config_one, "first 1")
        self.assertEqual(self.config.format('{1}', 'config'), config_one)
        """ get should mimic format behaviour, which we tested above """

        variables_three = self.loaded_variables.get('three')
        self.assertEqual(variables_three, 3)
        self.assertEqual(self.loaded_config.format("{variables:three}"), variables_three)
        self.assertEqual(self.loaded_config.format(" {variables:three} "), " {} ".format(variables_three))
        """ test templating for different sources """

        self.assertEqual(self.loaded_config.get('20'), "default")
        self.assertEqual(self.loaded_config.get('22'), "megadefault")
        self.assertEqual(self.loaded_config.get('23'), variables_three)
        """ test the templateing on get() """

    def test_missing_raises_exceptions(self):
        """ missing values should raise exceptions """

        with self.assertRaises(KeyError):
            self.config.format("{does.not.exist}", 'config')
            """ test template fails on missing keys in config format """

        with self.assertRaises(Exception):
            self.loaded_config.format("{does.not.exist}")
            """ test template fails on missing keys in loaded format """

        with self.assertRaises(Exception):
            self.loaded_config.get('21')
            """ test template fails on missing keys in get """

    def test_deep_formatting(self):
        """ test that formatting is applied deeply """

        config_fourty = self.loaded_config.get('40')
        self.assertEqual(config_fourty, "deep 40")

        self.assertEqual(self.loaded_config.get('41'), config_fourty)
        self.assertEqual(self.loaded_config.get('42')['1'], config_fourty)
        self.assertEqual(self.loaded_config.get('42')['2']['1'], config_fourty)

    def test_file_templating(self):
        """ test that file contents can be used as a formatting target """

        replace_config = self.config.load('replace')
        replace_json = replace_config.get(LOADED_KEY_ROOT)

        self.assertIsNotNone(replace_json)

        replace_path = replace_config.format('{paths:replace}')
        self.assertIsNotNone(replace_path)

        replace_full_path = os.path.join(replace_path, 'replace.json')
        with open(replace_full_path) as f:
            replace_json = json.load(f)
        self.assertIsNotNone(replace_json)

        # confirm that a file replace full match works
        self.assertEqual(replace_json, replace_config.format('[file:' + replace_full_path + ']'))
        self.assertEqual(
            ' {} '.format(
                json.dumps(replace_json)), replace_config.format(
                ' [file:' + replace_full_path + '] '))
        # confirm mixed templating
        self.assertEqual(replace_json, replace_config.format('[file:{paths:replace}/replace.json]'))



"""

Configerus Plugins

Typing and Factory decoration for configerus plugins.

"""
import logging
from enum import Enum, unique

logger = logging.getLogger('configerus.plugin')


@unique
class Type(Enum):
    """ Enumerator to match plugin types to plugin labels """
    SOURCE = "configerus.plugin.configsource"
    """ A Config source handler """
    FORMATTER = "configerus.plugin.formatter"
    """ A string formatter plugin """
    VALIDATOR = "configerus.plugin.validator"
    """ A data result validator plugin """


class Factory():
    """ Python decorator class for configuerus Plugin factories

    This class should be used to decorate any function which is meant to be a
    factory for plugins.

    If you are writing a plugin factory, decorate it with this class, provide
    the factory type and id values, and then the factory will be avaialble to
    other code.

    If you are trying to get an instance of a plugin, then create an instance of
    this class and use the create() method
    """

    registry = {}
    """ A static list of all of the registered factory functions """

    def __init__(self, type: Type, plugin_id: str):
        """ register the decoration """
        self.plugin_id = plugin_id
        self.type = type

        if not self.type.value in self.registry:
            self.registry[self.type.value] = {}

    def __call__(self, func):
        """ call the decorated function

        Returns:

        wrapped function(config: Config)
        """
        def wrapper(config, instance_id: str):
            logger.debug("plugin factory exec: %s:%s", self.type.value, self.plugin_id)
            return func(config=config, instance_id=instance_id)

        logger.debug("Plugin factory registered `%s:%s`", self.type.value, self.plugin_id)
        self.registry[self.type.value][self.plugin_id] = wrapper
        return wrapper

    def create(self, config, instance_id: str):
        """ Get an instance of a plugin as created by the decorated """
        try:
            factory = self.registry[self.type.value][self.plugin_id]
        except KeyError:
            raise NotImplementedError(
                "Configerus Plugin instance '{}:{}' has not been registered.".format(
                    self.type.value, self.plugin_id))
        # except Exception as e:
        #     raise Exception("Could not create Plugin instance '{}:{}' as the plugin factory produced an exception".format(self.type.value, self.plugin_id)) from e

        return factory(config=config, instance_id=instance_id)


CONFIGERUS_FORMATTER_TYPE = Type.FORMATTER
""" Short cut to the config source plugin type enum """


class FormatFactory(Factory):
    """ Decoration class for registering a config source plugin

    This decorator is just a shortcut to allow skipping identifying the plugin
    type using the core factory decorator class.

    All we do is extend the core Factory plugin and pass in the plugin type
    for formatters.
    """

    def __init__(self, plugin_id: str):
        """ register the decoration """
        super().__init__(CONFIGERUS_FORMATTER_TYPE, plugin_id)


CONFIGERUS_SOURCE_TYPE = Type.SOURCE
""" Short cut to the config source plugin type enum """


class SourceFactory(Factory):
    """ Decoration class for registering a config source plugin

    This decorator is just a shortcut to allow skipping identifying the plugin
    type using the core factory decorator class.

    All we do is extend the core Factory plugin and pass in the plugin type
    for formatters.
    """

    def __init__(self, plugin_id: str):
        """ register the decoration """
        super().__init__(CONFIGERUS_SOURCE_TYPE, plugin_id)


CONFIGERUS_VALIDATOR_TYPE = Type.VALIDATOR
""" Short cut to the validate plugin type enum """


class ValidatorFactory(Factory):
    """ Decoration class for registering a validate plugin

    This decorator is just a shortcut to allow skipping identifying the plugin
    type using the core factory decorator class.

    All we do is extend the core Factory plugin and pass in the plugin type
    for formatters.
    """

    def __init__(self, plugin_id: str):
        """ register the decoration """
        super().__init__(CONFIGERUS_VALIDATOR_TYPE, plugin_id)



"""

Test .sharted functions

Simple tests that ensure that the .shared functionality works

"""

import unittest
import logging

import configerus.shared as shared

logger = logging.getLogger("shared")


class BasicConstruct(unittest.TestCase):

    def test_1_tree_reduce(self):
        """ can we reduce trees to lists """
        tree = ['1', '2', '3']
        self.assertEqual(shared.tree_reduce(tree), tree)

        tree = ['1', '2', ['3', '4']]
        flat = ['1', '2', '3', '4']
        self.assertEqual(shared.tree_reduce(tree, glue='.'), flat)

        tree = ['1', '2', ['3', '4'], [['5'], '6', '7', [[['8']]]]]
        flat = ['1', '2', '3', '4', '5', '6', '7', '8']
        self.assertEqual(shared.tree_reduce(tree, glue='.'), flat)

    def test_2_tree_reduce_ignore(self):
        """ test the tree_reduce respects ignore """
        tree = ['1', '2', ['3', '4'], [['5'], '6', '7', [[['8']]]]]
        ignore = ['2', '4', '7']
        flat = ['1', '3', '5', '6', '8']
        self.assertEqual(shared.tree_reduce(tree, glue='.', ignore=ignore), flat)

    def test_3_tree_reduce_real(self):
        """ some gets that failed once upon a time """
        tree = ['', 'validators']
        ignore = ['', '']
        flat = ['validators']
        self.assertEqual(shared.tree_reduce(tree, glue='.', ignore=ignore), flat)

        tree = [['backend.outputs', 'mke_cluster'], 'validators']
        ignore = ['', '']
        flat = ['backend', 'outputs', 'mke_cluster', 'validators']
        self.assertEqual(shared.tree_reduce(tree, glue='.', ignore=ignore), flat)

    def test_4_tree_get(self):
        """ does tree_get() work with all sorts of messy keys """

        tree = {
            '1': {
                '2': {
                    '3': {
                        '4': 'my value'
                    }
                }
            }
        }

        self.assertEqual(shared.tree_get(tree, '1.2.3'), {'4': 'my value'})
        self.assertEqual(shared.tree_get(tree, '1.2.3.4'), 'my value')
        self.assertEqual(shared.tree_get(tree, '1.2.3.4.'), 'my value')
        self.assertEqual(shared.tree_get(tree, '1.2.3.4'), 'my value')
        self.assertEqual(shared.tree_get(tree, ['1', '2', '3', '4']), 'my value')
        self.assertEqual(shared.tree_get(tree, ['1.2', '3.4']), 'my value')
        self.assertEqual(shared.tree_get(tree, ['1', ['2', '3.4']]), 'my value')

        self.assertEqual(shared.tree_get(tree, ['1', '2', '.', '', '3.', '.4']), 'my value')
        self.assertEqual(shared.tree_get(tree, [['.', ''], '1.2.3.4', '..', ['.', ['.']]]), 'my value')

    def test_5_tree_merge(self):

        source = {
            '1': "source 1",
            '2': "source 2",
            '4': "source 4",
            '5': {
                '1': "source 5.1"
            }
        }
        destination = {
            '3': {
                '1': "destination 3.1",
                '2': {
                    '1': "destination 3.2.1",
                    '2': "destination 3.2.2"
                }
            },
            '4': {
                '1': "third 4.1"
            },
            '5': "third 5"
        }

        merged = shared.tree_merge(source, destination)

        self.assertIsNotNone(merged)
        self.assertEqual(shared.tree_get(merged, '3.1'), "destination 3.1")
        self.assertEqual(shared.tree_get(merged, '4'), "source 4")
        self.assertEqual(shared.tree_get(merged, '5.1'), "source 5.1")

        third = {}

    def test_get_listindexes(self):
        """ can get reach into lists using numberic indexes """

        test_list = [
            '1.2 0',
            '1.2 1',
            '1.2 2',
        ]
        wrapper_dict = {'2': test_list}
        tree = {
            '1': wrapper_dict
        }

        self.assertEqual(shared.tree_get(tree, '1'), wrapper_dict)
        self.assertEqual(shared.tree_get(tree, '1.2'), test_list)
        self.assertEqual(shared.tree_get(tree, '1.2.0'), test_list[0])
        self.assertEqual(shared.tree_get(tree, '1.2.2'), test_list[2])

        with self.assertRaises(IndexError):
            shared.tree_get(tree, '1.2.5')

        with self.assertRaises(ValueError):
            shared.tree_get(tree, '1.2.string')




from configerus.config import Config
from configerus.plugin import FormatFactory

from .format import ConfigFormatGetPlugin

PLUGIN_ID_FORMAT_GET = 'get'
""" Format plugin_id for the configerus get format plugin """
PLUGIN_PRIORITY_FORMAT_GET_PRIORITY = 80
""" bootstrapping get formatter plugin priority """


@FormatFactory(plugin_id=PLUGIN_ID_FORMAT_GET)
def plugin_factory_format_get(config: Config, instance_id: str = ''):
    """ create an format plugin which replaces from config.get """
    return ConfigFormatGetPlugin(config, instance_id)


def configerus_bootstrap(config: Config):
    """ Bootstrap a config object by adding our formatter """
    config.add_formatter(PLUGIN_ID_FORMAT_GET, priority=PLUGIN_PRIORITY_FORMAT_GET_PRIORITY)



"""

Test config's copy()

Here we test that the Config object is able to make copies of itself, and that
changing those copies doesn't affect originals.

"""

from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT
import configerus
import unittest
import logging

logging.basicConfig()
logger = logging.getLogger('config_copy')
logger.setLevel(level=logging.INFO)


""" TESTS """


class ConfigBehaviour(unittest.TestCase):

    def test_copy_safety(self):
        """ Test that config copy allows overloads and doesn't modify the source """

        logger.debug("Building empty config object")
        config = configerus.new_config()

        config.add_source(PLUGIN_ID_SOURCE_DICT, 'orig', 80).set_data({
            'copy': {
                'one': 'orig 1'
            }
        })

        config_copy_orig = config.load('copy')

        copy1 = config.copy()
        copy1.add_source(PLUGIN_ID_SOURCE_DICT, 'copy1', 81).set_data({
            'copy': {
                'one': 'copy1 1',
                'two': 'copy1 2'
            }
        })
        copy2 = config.copy()
        copy2.add_source(PLUGIN_ID_SOURCE_DICT, 'copy2', 82).set_data({
            'copy': {
                'one': 'copy2 1',
                'two': 'copy2 2'
            }
        })

        config_copy_late = config.load('copy')
        config1_copy = copy1.load('copy')
        config2_copy = copy2.load('copy')

        logger.debug('LOADED: AFTER: orig: %s', config_copy_orig.data)
        logger.debug('LOADED: AFTER: 1   : %s', config1_copy.data)
        logger.debug('LOADED: AFTER: 2   : %s', config2_copy.data)
        logger.debug('LOADED: AFTER: late: %s', config_copy_late.data)

        # check original values
        self.assertEqual(config_copy_orig.get('one'), 'orig 1')
        self.assertIsNone(config_copy_orig.get('two'))
        # check that copied config didn't modify original
        self.assertEqual(config_copy_orig.get('one'), config_copy_late.get('one'))
        self.assertEqual(config_copy_orig.get('two'), config_copy_late.get('two'))

        self.assertEqual(config1_copy.get('one'), 'copy1 1')
        self.assertEqual(config1_copy.get('two'), 'copy1 2')
